"""Contains a class for logic of about dialog."""

from PyQt5 import QtWidgets

from meggie.mainwindow.dialogs.aboutDialogUi import Ui_Dialog
from meggie.utilities.filemanager import get_distribution_version


class AboutDialog(QtWidgets.QDialog):
    """Contains the main logic for about dialog."""

    def __init__(self, parent):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)

        version = get_distribution_version("meggie")

        self.ui.lineEditVersion.setText(str(version))

    def on_pushButtonClose_clicked(self, checked=None):
        if checked is None:
            return
        self.close()



"""Contains spectrum creation logic."""

from collections import OrderedDict

import mne

import numpy as np

from meggie.datatypes.spectrum.spectrum import Spectrum

from meggie.utilities.threading import threaded
from meggie.utilities.events import get_raw_blocks_from_intervals


@threaded
def create_power_spectrum(subject, spectrum_name, intervals, fmin, fmax, nfft, overlap):
    """Creates a power spectrum item."""

    # get raw objects organized with average groups as keys
    ival_times, raw_block_groups = get_raw_blocks_from_intervals(subject, intervals)

    raw = subject.get_raw()
    picks = mne.pick_types(raw.info, meg=True, eeg=True, exclude="bads")

    # remove zero channels from picks
    zero_idxs = []
    for idx, row in enumerate(raw._data):
        if np.all(row == 0):
            zero_idxs.append(idx)
    picks = [pick for pick in picks if pick not in zero_idxs]

    # compute psd's
    psd_groups = OrderedDict()
    for key, raw_blocks in raw_block_groups.items():
        for raw_block in raw_blocks:
            length = len(raw_block.times)

            # create spectrum using mne's new api.
            # However, continue as before and convert
            # to plain freqs and data arrays.
            mne_spectrum = raw_block.compute_psd(
                method="welch",
                fmin=fmin,
                fmax=fmax,
                n_fft=nfft,
                n_overlap=overlap,
                picks=picks,
                proj=True,
            )
            psds = mne_spectrum.get_data()
            freqs = mne_spectrum.freqs

            if key not in psd_groups:
                psd_groups[key] = []

            psd_groups[key].append((psds, freqs, length))

    for psd_list in psd_groups.values():
        freqs = psd_list[0][1]
        break

    psds = []
    for psd_list in psd_groups.values():
        # do a weighted (raw block lengths as weights) average of psds inside a
        # group
        weights = np.array([length for psds_, freqs, length in psd_list])
        weights = weights.astype(float) / np.sum(weights)
        psd = np.average(
            [psds_ for psds_, freqs, length in psd_list], weights=weights, axis=0
        )
        psds.append(psd)

    info = mne.pick_info(raw.info, sel=picks)
    psd_data = dict(zip(psd_groups.keys(), psds))

    params = {}
    params["fmin"] = fmin
    params["fmax"] = fmax
    params["nfft"] = nfft
    params["overlap"] = overlap
    params["conditions"] = [elem for elem in psd_groups.keys()]
    params["intervals"] = ival_times

    spectrum = Spectrum(
        spectrum_name, subject.spectrum_directory, params, psd_data, freqs, info
    )

    spectrum.save_content()
    subject.add(spectrum, "spectrum")



"""Contains save evoked action handling."""

import os

from PyQt5 import QtWidgets

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.validators import assert_arrays_same
from meggie.utilities.filemanager import homepath

from meggie.mainwindow.dynamic import Action

from meggie.actions.evoked_save.controller.evoked import save_channel_averages
from meggie.actions.evoked_save.controller.evoked import save_all_channels

from meggie.utilities.dialogs.outputOptionsMain import OutputOptions


class SaveEvoked(Action):
    """Saved evoked items to csv files"""

    def run(self, params={}):
        try:
            selected_name = self.data["outputs"]["evoked"][0]
        except IndexError:
            return

        # validate times
        time_arrays = []
        for subject in self.experiment.subjects.values():
            evoked = subject.evoked.get(selected_name)
            if not evoked:
                continue
            for mne_evoked in evoked.content.values():
                time_arrays.append(mne_evoked.times)

        assert_arrays_same(time_arrays, "Times do not match")

        def option_handler(selected_option):

            default_filename = (
                selected_name + "_all_subjects_channel_averages_evoked.csv"
                if selected_option == "channel_averages"
                else selected_name + "_all_subjects_all_channels_evoked.csv"
            )
            filepath, _ = QtWidgets.QFileDialog.getSaveFileName(
                self.window,
                "Save Evoked to CSV",
                os.path.join(homepath(), default_filename),
                "CSV Files (*.csv);;All Files (*)",
            )
            if not filepath:
                return

            params = {
                "name": selected_name,
                "output_option": selected_option,
                "channel_groups": self.experiment.channel_groups,
                "filepath": filepath,
            }

            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = OutputOptions(self.window, handler=option_handler)
        dialog.show()

    def handler(self, subject, params):
        """ """
        if params["output_option"] == "channel_averages":
            save_channel_averages(
                self.experiment,
                params["name"],
                params["channel_groups"],
                params["filepath"],
                do_meanwhile=self.window.update_ui,
            )
        else:
            save_all_channels(
                self.experiment,
                params["name"],
                params["filepath"],
                do_meanwhile=self.window.update_ui,
            )



import mne
from meggie.experiment import initialize_new_experiment
from meggie.datatypes.epochs.epochs import Epochs


def create_limo_experiment(name, author, prefs, set_previous_experiment=True):
    """Generate experiment based on the open limo data.

    Parameters
    ----------
    name : str
        Name of the experiment.
    author : str
        Name of the author.
    prefs : meggie.mainwindow.preferences.PreferencesHandler
        A preferences object.

    Returns
    -------
    meggie.experiment.Experiment
        The new experiment.
    """

    experiment = initialize_new_experiment(name, author, prefs, set_previous_experiment)

    for idx in range(0, 18):
        limo_data = mne.datasets.limo.load_data(idx + 1, update_path=False)

        subject_name = f"subject_{str(idx + 1).zfill(2)}-raw"

        # create a subject without raw
        subject = experiment.create_subject(subject_name, None)

        epochs_directory = subject.epochs_directory

        mne_epochs_1 = limo_data["Face/A"]
        params = {
            "tmin": mne_epochs_1.times[0],
            "tmax": mne_epochs_1.times[-1],
            "bstart": mne_epochs_1.times[0],
            "bend": 0.0,
        }
        epochs_1 = Epochs("FaceA", epochs_directory, params, content=mne_epochs_1)
        epochs_1.save_content()
        subject.add(epochs_1, "epochs")

        mne_epochs_2 = limo_data["Face/B"]
        params = {
            "tmin": mne_epochs_2.times[0],
            "tmax": mne_epochs_2.times[-1],
            "bstart": mne_epochs_2.times[0],
            "bend": 0.0,
        }
        epochs_2 = Epochs("FaceB", epochs_directory, params, content=mne_epochs_2)
        epochs_2.save_content()
        subject.add(epochs_2, "epochs")

    return experiment


def create_somato_experiment(name, author, prefs, set_previous_experiment=True):
    """Generate one-subject experiment based on the mne sample data.

    Parameters
    ----------
    name : str
        Name of the experiment.
    author : str
        Name of the author.
    prefs : meggie.mainwindow.preferences.PreferencesHandler
        A preferences object.

    Returns
    -------
    meggie.experiment.Experiment
        The new experiment.
    """

    experiment = initialize_new_experiment(name, author, prefs, set_previous_experiment)

    mne.datasets.somato.data_path()

    somato_data = (
        str(mne.datasets.somato.data_path()) + "/sub-01/meg/sub-01_task-somato_meg.fif"
    )

    experiment.create_subject("sub-01", somato_data)

    return experiment


def create_sample_experiment(name, author, prefs, set_previous_experiment=True):
    """Generate one-subject experiment based on the mne somato data.

    Parameters
    ----------
    name : str
        Name of the experiment.
    author : str
        Name of the author.
    prefs : meggie.mainwindow.preferences.PreferencesHandler
        A preferences object.

    Returns
    -------
    meggie.experiment.Experiment
        The new experiment.
    """

    experiment = initialize_new_experiment(name, author, prefs, set_previous_experiment)

    sample_data = (
        str(mne.datasets.sample.data_path()) + "/MEG/sample/sample_audvis_raw.fif"
    )
    experiment.create_subject("sample_audvis_raw", sample_data)

    return experiment


def get_open_datasets():
    return {
        "limo": {"title": "MNE LIMO Dataset", "constructor": create_limo_experiment},
        "somato": {
            "title": "MNE Somato Dataset",
            "constructor": create_somato_experiment,
        },
        "sample": {
            "title": "MNE Sample Dataset",
            "constructor": create_sample_experiment,
        },
    }



"""Contains implementation for delete tfr from all"""

import logging

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class DeleteTFRFromAll(Action):
    """Deletes TFR of selected name from all subjects"""

    def run(self, params={}):

        try:
            selected_name = self.data["outputs"]["tfr"][0]
        except IndexError:
            return

        for subject in self.experiment.subjects.values():
            if selected_name in subject.tfr:
                try:
                    self.handler(subject, {"name": selected_name})
                except Exception:
                    logging.getLogger("ui_logger").exception("")
                    logging.getLogger("ui_logger").warning(
                        "Could not remove TFR for " + subject.name
                    )

        self.window.initialize_ui()

    @subject_action
    def handler(self, subject, params):
        subject.remove(params["name"], "tfr")
        self.experiment.save_experiment_settings()



"""Contains the class for main window logic."""

import os
import sys
import logging
import warnings

from pythonjsonlogger import jsonlogger

from PyQt5.QtWidgets import QApplication
from PyQt5 import QtWidgets
from PyQt5 import QtCore

from meggie.mainwindow.dynamic import construct_tabs

from meggie.mainwindow.preferences import PreferencesHandler

from meggie.mainWindowUi import Ui_MainWindow
from meggie.experiment import open_existing_experiment

from meggie.utilities.threading import threaded
from meggie.utilities.messaging import questionbox
from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox

from meggie.mainwindow.dialogs.actionDialogMain import ActionDialog
from meggie.mainwindow.dialogs.aboutDialogMain import AboutDialog
from meggie.mainwindow.dialogs.preferencesDialogMain import PreferencesDialog
from meggie.mainwindow.dialogs.channelGroupsDialogMain import ChannelGroupsDialog
from meggie.mainwindow.dialogs.pipelineDialogMain import PipelineDialog
from meggie.mainwindow.dialogs.addSubjectDialogMain import AddSubjectDialog
from meggie.mainwindow.dialogs.createExperimentDialogMain import CreateExperimentDialog


class MainWindow(QtWidgets.QMainWindow):
    """Contains the main window logic and stores the experiment."""

    def __init__(self, application):
        QtWidgets.QMainWindow.__init__(self)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # set default ratio of splitter
        self.ui.splitterTopBottom.setStretchFactor(0, 2)
        self.ui.splitterTopBottom.setStretchFactor(1, 1)

        self.experiment = None

        self._setup_loggers()

        # Direct output to console
        if not sys.argv[-1] == "debug":
            self._direct_output()

        # For storing and handling program wide preferences.
        self.prefs = PreferencesHandler()

        auto_load = self.prefs.auto_load_last_open_experiment
        previous_name = self.prefs.previous_experiment_name
        if auto_load and previous_name:
            try:
                exp = open_existing_experiment(self.prefs)
                self.experiment = exp
                self.prefs.previous_experiment_name = exp.path
                self.prefs.write_preferences_to_disk()
                logging.getLogger("ui_logger").info("Opening experiment " + exp.path)
            except Exception as exc:
                self.prefs.previous_experiment_name = ""
                exc_messagebox(self, exc)

        self.reconstruct_tabs()
        self.initialize_ui()

    def on_actionQuit_triggered(self, checked=None):
        if checked is None:
            return
        self.close()

    def on_actionCreateExperiment_triggered(self, checked=None):
        if checked is None:
            return

        if not self.prefs.workspace:
            messagebox(
                self,
                "Please set up a workspace "
                "(a place where experiments are stored) in preferences "
                "before creating experiments",
            )
            dialog = PreferencesDialog(self)
            dialog.show()
        else:
            dialog = CreateExperimentDialog(self)
            dialog.show()

    def on_actionOpenExperiment_triggered(self, checked=None):
        if checked is None:
            return

        directory = self.prefs.workspace
        path = QtCore.QDir.toNativeSeparators(
            str(
                QtWidgets.QFileDialog.getExistingDirectory(
                    self, "Select experiment directory", directory
                )
            )
        )

        if path == "":
            return

        logging.getLogger("ui_logger").info("Opening experiment " + path)

        try:
            exp = open_existing_experiment(self.prefs, path=path)
            self.experiment = exp
            self.prefs.previous_experiment_name = exp.path
            self.prefs.write_preferences_to_disk()
            self.reconstruct_tabs()
            self.initialize_ui()
        except Exception as exc:
            exc_messagebox(self, exc)

    def on_pushButtonAddSubjects_clicked(self, checked=None):
        if checked is None:
            return

        # Check that we have an experiment that we can add a subject to
        if not self.experiment:
            msg = (
                "No active experiment to add a subject to. Load an "
                "experiment or make a new one, then try again."
            )
            messagebox(self, msg)
            return

        dialog = AddSubjectDialog(self)
        dialog.show()

    def on_pushButtonRemoveSubject_clicked(self, checked=None):
        if checked is None:
            return

        selIndexes = self.ui.listWidgetSubjects.selectedIndexes()

        if selIndexes == []:
            return

        def handler(accepted):
            if not accepted:
                return

            n_successful = 0
            for index in selIndexes:
                subject_name = index.data()
                try:
                    self.experiment.remove_subject(subject_name)
                    n_successful += 1
                except Exception:
                    logging.getLogger("ui_logger").exception("")

            try:
                self.experiment.save_experiment_settings()
            except Exception as exc:
                exc_messagebox(self, exc)
                return

            n_total = len(selIndexes)

            if n_successful != n_total:
                message = (
                    "Could not remove all subjects completely. "
                    "Please check console below for details."
                )
                messagebox(self, message)

            self.initialize_ui()

        questionbox(
            self,
            "Permanently remove the selected subjects and the related files?",
            handler,
        )

    def on_actionActions_triggered(self, checked=None):
        if checked is None:
            return

        if not self.experiment:
            message = "Please open an experiment first."
            messagebox(self, message)
            return

        dialog = ActionDialog(self, self.experiment)
        dialog.show()

    def on_actionPreferences_triggered(self, checked=None):
        if checked is None:
            return

        dialog = PreferencesDialog(self)
        dialog.show()

    def on_actionAbout_triggered(self, checked=None):
        if checked is None:
            return

        dialog = AboutDialog(self)
        dialog.show()

    def on_pushButtonChannelGroups_clicked(self, checked=None):
        if checked is None:
            return

        if not self.experiment:
            return

        dialog = ChannelGroupsDialog(self)
        dialog.show()

    def on_pushButtonPipelines_clicked(self, checked=None):
        if checked is None:
            return

        if not self.experiment:
            return

        dialog = PipelineDialog(self, self.prefs)
        dialog.show()

    def on_pushButtonActivateSubject_clicked(self, checked=None):
        if checked is None:
            return

        items = self.ui.listWidgetSubjects.selectedItems()
        if not items:
            return

        subject_name = items[0].text()

        if self.experiment.active_subject:
            if subject_name == self.experiment.active_subject.name:
                return

        previous_subject = self.experiment.active_subject
        try:

            @threaded
            def activate(subject_name):
                self.experiment.activate_subject(subject_name)

            activate(subject_name, do_meanwhile=self.update_ui)

            self.reconstruct_tabs()

        except Exception:
            self.experiment.active_subject = None
            messagebox(self, "Could not activate the subject.")

            if previous_subject:
                message = "Couldn't activate the subject, resuming to previous one."
                logging.getLogger("ui_logger").info(message)
                self.experiment.activate_subject(previous_subject.name)

        self.initialize_ui()

    def update_ui(self):
        """Used for keeping the ui responsive when threading."""
        QApplication.processEvents()

    def reconstruct_tabs(self):
        """Reconstructs the tabs."""

        include_eeg = True
        has_raw = True
        if self.experiment:
            active_subject = self.experiment.active_subject
            if active_subject:
                has_raw = active_subject.has_raw
                if has_raw:
                    include_eeg = active_subject.has_eeg

        if self.experiment:
            selected_pipeline = self.experiment.selected_pipeline
        else:
            selected_pipeline = "classic"

        try:
            self.tabs = construct_tabs(
                selected_pipeline,
                self,
                self.prefs,
                include_eeg=include_eeg,
                has_raw=has_raw,
            )
        except Exception as exc:
            self.tabs = []
            exc_messagebox(self, exc)

    def initialize_ui(self):
        """Initializes the main window UI view.

        Often used if the underlying experiment changes.
        """
        self._update_tabs()
        self._setup_loggers()

        self.ui.listWidgetSubjects.clear()

        if not self.experiment:
            self.setWindowTitle("Meggie")
            return

        if self.experiment.name:
            self.ui.labelExperimentNameValue.setText(self.experiment.name)

        if self.experiment.author:
            self.ui.labelExperimentAuthorValue.setText(self.experiment.author)
        else:
            self.ui.labelExperimentAuthorValue.setText("")

        self.setWindowTitle("Meggie - " + self.experiment.name)

        self._populate_subject_list()

    def _populate_subject_list(self):
        active_subject_name = None
        if self.experiment and self.experiment.active_subject:
            active_subject_name = self.experiment.active_subject.name

        for subject_name in sorted(self.experiment.subjects.keys()):
            item = QtWidgets.QListWidgetItem()
            item.setText(subject_name)
            if subject_name == active_subject_name:
                font = item.font()
                font.setBold(True)
                item.setFont(font)
            self.ui.listWidgetSubjects.addItem(item)

    def _update_tabs(self):
        current_tab = self.ui.tabWidget.currentIndex()
        while self.ui.tabWidget.count() > 0:
            self.ui.tabWidget.removeTab(0)

        for tab_idx, tab in enumerate(self.tabs):
            self.ui.tabWidget.insertTab(tab_idx + 1, tab, tab.name)

        self.ui.tabWidget.setCurrentIndex(current_tab)

        for tab in self.tabs:
            tab.initialize_ui()

    def _direct_output(self):
        stdout_stream = EmittingStream(textWritten=self._normal_output_written)
        stdout_stream.orig_stream = sys.__stdout__
        stderr_stream = EmittingStream(textWritten=self._error_output_written)
        stderr_stream.orig_stream = sys.__stderr__

        sys.stdout = stdout_stream
        sys.stderr = stderr_stream

    def _normal_output_written(self, text):
        cursor = self.ui.textEditConsole.textCursor()
        cursor.movePosition(cursor.MoveOperation.End)
        cursor.insertText(text)
        self.ui.textEditConsole.setTextCursor(cursor)
        self.ui.textEditConsole.ensureCursorVisible()

    def _error_output_written(self, text):
        cursor = self.ui.textEditConsole.textCursor()
        cursor.movePosition(cursor.MoveOperation.End)
        cursor.insertText(text)
        self.ui.textEditConsole.setTextCursor(cursor)
        self.ui.textEditConsole.ensureCursorVisible()

    def _setup_loggers(self):
        # hide warnings-module warnings,
        # most of these are still contained
        # in mne-level logging
        warnings.simplefilter("ignore")

        logging.getLogger().setLevel(logging.DEBUG)

        logger_error_message = (
            "Could not setup loggers because of missing "
            "permissions. The whole experiment folder "
            "should have write permissions."
        )

        # setup logger for informative messages
        ui_logger = logging.getLogger("ui_logger")
        ui_logger.propagate = False
        formatter = logging.Formatter(
            "Meggie: %(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        ui_logger.handlers = []
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        stream_handler.setLevel("INFO")
        ui_logger.addHandler(stream_handler)

        # setup logger for mne error messages
        mne_logger = logging.getLogger("mne")
        mne_logger.propagate = False
        formatter = logging.Formatter(
            "MNE: %(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )

        mne_logger.handlers = []
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        stream_handler.setLevel("ERROR")
        mne_logger.addHandler(stream_handler)

        # setup action logger
        action_logger = logging.getLogger("action_logger")
        action_logger.propagate = False
        action_logger.handlers = []

        # setup file handler
        if self.experiment:
            try:
                logfile = os.path.join(self.experiment.path, "actions.log")
                file_handler = logging.FileHandler(logfile)
                file_handler.setLevel("INFO")

                formatter = jsonlogger.JsonFormatter(timestamp=True)
                file_handler.setFormatter(formatter)
                action_logger.addHandler(file_handler)
            except PermissionError:
                raise Exception(logger_error_message)


class EmittingStream(QtCore.QObject):
    """Helper class for console."""

    textWritten = QtCore.pyqtSignal(str)

    def write(self, text):
        self.orig_stream.write(str(text))
        self.textWritten.emit(str(text))

    def flush(self):
        pass

    def fileno(self):
        return self.orig_stream.fileno()


def main():
    # Create the window.
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow(app)
    window.showMaximized()
    sys.exit(app.exec())



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_rereference import Rereference
from meggie.actions.raw_rereference.dialogs.rereferencingDialogMain import (
    RereferencingDialog,
)


class TestRawRereference(BaseTestAction):
    def test_raw_rereference(self):

        self.run_action(
            action_name="raw_rereference",
            handler=Rereference,
            patch_paths=[
                "meggie.actions.raw_rereference.dialogs.rereferencingDialogMain"
            ],
        )
        dialog = self.find_dialog(RereferencingDialog)
        dialog.accept()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.evoked_create import CreateEvoked
from meggie.utilities.dialogs.simpleDialogMain import SimpleDialog


class TestEvokedCreate(BaseTestAction):
    def test_evoked_create(self):

        data = {"inputs": {"epochs": ["Epochs"]}}

        self.run_action(
            action_name="evoked_create",
            handler=CreateEvoked,
            data=data,
            patch_paths=["meggie.utilities.dialogs.simpleDialogMain"],
        )
        dialog = self.find_dialog(SimpleDialog)
        dialog.accept()



"""Contains implementation for raw rereference"""

from meggie.utilities.threading import threaded

from meggie.actions.raw_rereference.dialogs.rereferencingDialogMain import (
    RereferencingDialog,
)

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class Rereference(Action):
    """Shows a dialog and then allows rereferencing eeg data."""

    def run(self, params={}):
        rereference_dialog = RereferencingDialog(
            self.window, self.experiment, self.handler
        )
        rereference_dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """

        @threaded
        def rereference_fun():
            raw = subject.get_raw()
            raw = raw.set_eeg_reference(
                ref_channels=params["selection"],
                projection=False,
            )

        rereference_fun(do_meanwhile=self.window.update_ui)
        subject.rereferenced = True
        subject.save()
        self.experiment.save_experiment_settings()



"""Contains a class for logic of add subject dialog."""

import os
import mne
import logging

from PyQt5 import QtWidgets
from PyQt5 import QtCore

from meggie.mainwindow.dialogs.addSubjectDialogUi import Ui_AddSubject

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.threading import threaded
from meggie.utilities.names import next_available_name
from meggie.utilities.filemanager import get_supported_formats
from meggie.utilities.filemanager import homepath


class AddSubjectDialog(QtWidgets.QDialog):
    """Contains the logic for add subject dialog."""

    def __init__(self, parent):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_AddSubject()
        self.ui.setupUi(self)

        self.parent = parent

    def accept(self):
        n_successful = 0
        for i in range(self.ui.listWidgetFileNames.count()):
            item = self.ui.listWidgetFileNames.item(i)
            raw_path = item.text()
            basename = os.path.basename(raw_path)
            subject_name = basename.split(".")[0]

            experiment = self.parent.experiment
            old_names = experiment.subjects.keys()

            try:
                subject_name = next_available_name(old_names, subject_name)

                @threaded
                def _create_subject():
                    experiment.create_subject(subject_name, raw_path)

                _create_subject(do_meanwhile=self.parent.update_ui)
                n_successful += 1

            except Exception:
                logging.getLogger("ui_logger").exception("")

        try:
            self.parent.experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        n_total = self.ui.listWidgetFileNames.count()

        if n_total != n_successful:
            message = (
                "Only {0} / {1} subjects added successfully. "
                "Please check console below for details."
            )
            messagebox(self.parent, message.format(n_successful, n_total))

        message = ("{0} / {1} subjects added successfully.").format(
            n_successful, n_total
        )
        logging.getLogger("ui_logger").info(message)

        self.parent.initialize_ui()
        self.close()

    def on_pushButtonSample_clicked(self, checked=None):
        if checked is None:
            return

        experiment = self.parent.experiment
        old_names = experiment.subjects.keys()
        subject_name = next_available_name(old_names, "sample_audvis_raw")

        @threaded
        def _create_subject():
            raw_path = os.path.join(
                str(mne.datasets.sample.data_path()),
                "MEG",
                "sample",
                "sample_audvis_raw.fif",
            )
            experiment.create_subject(subject_name, raw_path)

        try:
            _create_subject(do_meanwhile=self.parent.update_ui)
        except Exception:
            logging.getLogger("ui_logger").exception("")

        try:
            self.parent.experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def on_pushButtonBrowse_clicked(self, checked=None):
        if checked is None:
            return

        mne_supported = get_supported_formats()

        all_ext = [val[0] for val in mne_supported]

        all_items = []
        for ext, names in mne_supported:
            for name in names:
                all_items.append(name + " files (*" + ext + ")")

        filter_string = "All supported (*" + " *".join(all_ext) + ");" ";"
        filter_string = filter_string + ";" ";".join(all_items)

        self.fnames = QtWidgets.QFileDialog.getOpenFileNames(
            self,
            "Select one or more files to open.",
            homepath(),
            filter_string,
        )[0]

        if len(self.fnames) > 0:
            for name in self.fnames:
                name = QtCore.QDir.toNativeSeparators(name)
                item = QtWidgets.QListWidgetItem()
                item.setText(name)
                if (
                    len(
                        self.ui.listWidgetFileNames.findItems(
                            name, QtCore.Qt.MatchExactly
                        )
                    )
                    > 0
                ):
                    continue
                self.ui.listWidgetFileNames.addItem(item)

    def on_pushButtonRemove_clicked(self, checked=None):
        if checked is None:
            return
        for item in self.ui.listWidgetFileNames.selectedItems():
            i = self.ui.listWidgetFileNames.indexFromItem(item)
            row = i.row()
            self.ui.listWidgetFileNames.takeItem(row)

    def on_listWidgetFileNames_itemSelectionChanged(self):
        items = self.ui.listWidgetFileNames.selectedItems()
        if len(items) > 0:
            self.ui.pushButtonRemove.setEnabled(True)
        else:
            self.ui.pushButtonRemove.setEnabled(False)



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_resample import Resample
from meggie.actions.raw_resample.dialogs.resamplingDialogMain import (
    ResamplingDialog,
)


class TestRawResample(BaseTestAction):
    def test_raw_resample(self):

        self.run_action(
            action_name="raw_resample",
            handler=Resample,
            patch_paths=["meggie.actions.raw_resample.dialogs.resamplingDialogMain"],
        )
        dialog = self.find_dialog(ResamplingDialog)
        dialog.accept()



"""Contains controlling logic for the epochs implementation."""

import logging

import numpy as np

import meggie.utilities.filemanager as filemanager

from meggie.utilities.formats import format_floats
from meggie.utilities.channels import average_to_channel_groups
from meggie.utilities.threading import threaded


def _average_epochs_to_channel_groups(data, info, ch_names, channel_groups):
    """Averages epochs data to channel groups."""
    all_averaged_data = []
    for epoch_data in data:
        data_labels, averaged_data = average_to_channel_groups(
            epoch_data, info, ch_names, channel_groups
        )
        all_averaged_data.append(averaged_data)
    return data_labels, np.array(all_averaged_data)


@threaded
def save_all_channels(experiment, selected_name, path, bstart, bend):
    """Saves all channels of epochs item to a csv file."""
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        epochs = subject.epochs.get(selected_name)
        if not epochs:
            continue

        mne_epochs = epochs.content.copy()
        if bstart and bend:
            mne_epochs.apply_baseline((bstart, bend))

        column_names = format_floats(mne_epochs.times)

        for epoch_idx, epoch_data in enumerate(mne_epochs.get_data()):
            for ch_idx, ch_name in enumerate(mne_epochs.ch_names):
                if ch_name in mne_epochs.info["bads"]:
                    continue
                csv_data.append(epoch_data[ch_idx].tolist())

                row_desc = (subject.name, epoch_idx, ch_name)
                row_descs.append(row_desc)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger("ui_logger").info("Saved the csv file to " + path)


@threaded
def save_channel_averages(
    experiment, selected_name, channel_groups, path, bstart, bend
):
    """Saves channel averages of epochs item to a csv file."""
    column_names = []
    row_descs = []
    csv_data = []

    # accumulate csv contents
    for subject in experiment.subjects.values():
        epochs = subject.epochs.get(selected_name)
        if not epochs:
            continue

        mne_epochs = epochs.content.copy()
        mne_epochs.load_data()
        if bstart and bend:
            mne_epochs.apply_baseline((bstart, bend))

        mne_epochs.drop_channels(mne_epochs.info["bads"])

        data_labels, averaged_data = _average_epochs_to_channel_groups(
            mne_epochs.get_data(), mne_epochs.info, mne_epochs.ch_names, channel_groups
        )

        column_names = format_floats(mne_epochs.times)

        for epoch_idx, epoch_data in enumerate(averaged_data):
            for ch_idx, (ch_type, area) in enumerate(data_labels):
                csv_data.append(epoch_data[ch_idx].tolist())
                row_desc = (subject.name, epoch_idx, ch_type, area)
                row_descs.append(row_desc)

    filemanager.save_csv(path, csv_data, column_names, row_descs)
    logging.getLogger("ui_logger").info("Saved the csv file to " + path)



from meggie.utilities.testing import BaseTestAction
from meggie.actions.epochs_delete_from_all import DeleteEpochsFromAll


class TestEpochsDeleteFromAll(BaseTestAction):
    def test_epochs_delete_from_all(self):

        data = {"outputs": {"epochs": ["Epochs"]}}

        self.run_action(
            action_name="epochs_delete_from_all",
            handler=DeleteEpochsFromAll,
            data=data,
        )



"""Contains create tfr action handling."""

from meggie.utilities.names import next_available_name

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.tfr_create.dialogs.TFRDialogMain import TFRDialog
from meggie.actions.tfr_create.controller.tfr import create_tfr


class CreateTFR(Action):
    """Shows a TFR parameter dialog and then
    creates TFR"""

    def run(self, params={}):

        selected_names = self.data["inputs"]["epochs"]

        if not selected_names:
            return

        stem = "TFR"

        default_name = next_available_name(
            self.experiment.active_subject.tfr.keys(), stem
        )

        dialog = TFRDialog(
            self.experiment, self.window, selected_names, default_name, self.handler
        )
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """
        tfr_name = params["name"]
        epochs_names = params["conditions"]
        freqs = params["freqs"]
        decim = params["decim"]
        n_cycles = params["n_cycles"]
        subtract_evoked = params["subtract_evoked"]

        create_tfr(
            subject,
            tfr_name,
            epochs_names,
            freqs,
            decim,
            n_cycles,
            subtract_evoked,
            do_meanwhile=self.window.update_ui,
        )
        self.experiment.save_experiment_settings()



from meggie.utilities.plotting import color_cycle


def test_color_cycle():
    colors = color_cycle(30)
    assert isinstance(colors, list)
    assert len(colors) == 30
    assert len(set(colors)) == 8



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_plot import PlotRaw


class TestRawPlot(BaseTestAction):
    def test_raw_plot(self):

        # this might not call the handler as 'on_close' is not called
        self.run_action(
            action_name="raw_plot",
            handler=PlotRaw,
        )



"""Contains tfr group average action handling."""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.names import next_available_name

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.tfr_group_average.controller.tfr import group_average_tfr

from meggie.utilities.dialogs.groupSelectionDialogMain import GroupSelectionDialog


class GroupAverage(Action):
    """Creates group average items"""

    def run(self, params={}):
        """ """
        try:
            selected_name = self.data["outputs"]["tfr"][0]
        except IndexError:
            return

        name = next_available_name(
            self.experiment.active_subject.tfr.keys(), "group_" + selected_name
        )

        def group_handler(groups):
            params = {"based_on": selected_name, "name": name, "groups": groups}
            try:
                self.handler(self.experiment.active_subject, params)
                self.window.initialize_ui()
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = GroupSelectionDialog(
            self.experiment, self.window, handler=group_handler
        )
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """
        group_average_tfr(
            self.experiment,
            params["based_on"],
            params["groups"],
            params["name"],
            do_meanwhile=self.window.update_ui,
        )
        self.experiment.save_experiment_settings()



"""Contains controlling logic for the ICA."""

import mne

from meggie.utilities.compare import compare_raws
from meggie.utilities.plotting import set_figure_title


def compute_ica(raw, n_components, method, max_iter, random_state):
    """Computes ICA using MNE implementation."""

    ica = mne.preprocessing.ICA(
        n_components=n_components,
        method=method,
        max_iter=max_iter,
        random_state=random_state,
    )

    ica.fit(raw)
    return ica


def plot_topographies(ica, n_components):
    """Plots topographies from the ICA solution."""

    figs = ica.plot_components(title="")
    if not isinstance(figs, list):
        figs = [figs]

    for fig in figs:
        set_figure_title(fig, "ICA topographic maps")

    idx = 0
    for fig in figs:
        for ax in fig.get_axes():
            if idx > n_components:
                return

            ax.set_title("Component " + str(idx), fontsize=12)
            idx += 1


def plot_sources(raw, ica):
    """Plots sources of the ica solution."""
    sources = ica.get_sources(raw)
    sources.plot(title="ICA time courses")


def plot_properties(raw, ica, picks):
    """Plots properties for specific ICA components."""
    figs = ica.plot_properties(raw, picks)
    for fig in figs:
        set_figure_title(fig, "ICA properties")

    # fix the names
    idx = 0
    for fig in figs:
        for ax_idx, ax in enumerate(fig.get_axes()):
            if ax_idx == 0:
                ax.set_title("Component " + str(picks[idx]))
                idx += 1
            break


def plot_changes(raw, ica, indices):
    """Plot a raw comparison plot for ICA solution."""
    raw_removed = raw.copy()
    ica.apply(raw_removed, exclude=indices)
    compare_raws(raw, raw_removed)



"""Contains implementation for epochs plot"""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.plotting import set_figure_title
from meggie.utilities.plotting import get_figure_title
from meggie.mainwindow.dynamic import Action


class PlotEpochsImage(Action):
    """Opens MNE's image plot."""

    def run(self, params={}):

        try:
            selected_name = self.data["outputs"]["epochs"][0]
        except IndexError:
            return

        subject = self.experiment.active_subject
        try:
            self.handler(subject, {"name": selected_name})
        except Exception as exc:
            exc_messagebox(self.window, exc)

    def handler(self, subject, params):
        epochs = subject.epochs.get(params["name"])
        mne_epochs = epochs.content
        figs = mne_epochs.plot_image()
        for fig in figs:
            ch_type = get_figure_title(fig)
            title = "{0}_{1}".format(params["name"], ch_type)
            set_figure_title(fig, title)



"""Contains create evoked action handling."""

from meggie.utilities.names import next_available_name
from meggie.utilities.validators import assert_arrays_same
from meggie.utilities.threading import threaded

from meggie.datatypes.evoked.evoked import Evoked

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.utilities.dialogs.simpleDialogMain import SimpleDialog


class CreateEvoked(Action):
    """Allows averaging epochs to evoked"""

    def run(self, params={}):

        selected_names = self.data["inputs"]["epochs"]

        if not selected_names:
            return

        stem = "Evoked"

        default_name = next_available_name(
            self.experiment.active_subject.evoked.keys(), stem
        )

        def close_handle(subject, params):
            params["conditions"] = selected_names
            self.handler(subject, params)

        dialog = SimpleDialog(
            self.experiment,
            self.window,
            default_name,
            close_handle,
            title="Create evoked",
        )
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """
        # check that selected epochs have similar times structure
        time_arrays = []
        for name in params["conditions"]:
            epochs = subject.epochs.get(name)
            if epochs:
                time_arrays.append(epochs.content.times)
        assert_arrays_same(time_arrays)

        evokeds = {}
        for name in params["conditions"]:
            epochs = subject.epochs.get(name)
            if not epochs:
                raise KeyError("No epoch collection called " + str(name))
            mne_epochs = epochs.content

            @threaded
            def average():
                return mne_epochs.average()

            mne_evoked = average(do_meanwhile=self.window.update_ui)
            evokeds[name] = mne_evoked

        evoked_directory = subject.evoked_directory
        evoked = Evoked(params["name"], evoked_directory, params, content=evokeds)
        evoked.save_content()
        subject.add(evoked, "evoked")
        self.experiment.save_experiment_settings()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './meggie/meggie/actions/raw_resample/dialogs/resamplingDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_resamplingDialog(object):
    def setupUi(self, resamplingDialog):
        resamplingDialog.setObjectName("resamplingDialog")
        resamplingDialog.resize(406, 540)
        self.gridLayout = QtWidgets.QGridLayout(resamplingDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea = QtWidgets.QScrollArea(resamplingDialog)
        self.scrollArea.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setSizeAdjustPolicy(
            QtWidgets.QAbstractScrollArea.AdjustToContents
        )
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 386, 489))
        self.scrollAreaWidgetContents.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.groupBoxResample = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxResample.setObjectName("groupBoxResample")
        self.formLayout = QtWidgets.QFormLayout(self.groupBoxResample)
        self.formLayout.setObjectName("formLayout")
        self.labelCurrentRateHeading = QtWidgets.QLabel(self.groupBoxResample)
        self.labelCurrentRateHeading.setObjectName("labelCurrentRateHeading")
        self.formLayout.setWidget(
            0, QtWidgets.QFormLayout.LabelRole, self.labelCurrentRateHeading
        )
        self.labelCurrentRateValue = QtWidgets.QLabel(self.groupBoxResample)
        self.labelCurrentRateValue.setText("")
        self.labelCurrentRateValue.setObjectName("labelCurrentRateValue")
        self.formLayout.setWidget(
            0, QtWidgets.QFormLayout.FieldRole, self.labelCurrentRateValue
        )
        self.labelNewRateHeading = QtWidgets.QLabel(self.groupBoxResample)
        self.labelNewRateHeading.setObjectName("labelNewRateHeading")
        self.formLayout.setWidget(
            1, QtWidgets.QFormLayout.LabelRole, self.labelNewRateHeading
        )
        self.doubleSpinBoxNewRate = QtWidgets.QDoubleSpinBox(self.groupBoxResample)
        self.doubleSpinBoxNewRate.setMaximum(10000.0)
        self.doubleSpinBoxNewRate.setProperty("value", 100.0)
        self.doubleSpinBoxNewRate.setObjectName("doubleSpinBoxNewRate")
        self.formLayout.setWidget(
            1, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxNewRate
        )
        self.gridLayout_2.addWidget(self.groupBoxResample, 0, 0, 1, 1)
        self.groupBoxBatching = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxBatching.setObjectName("groupBoxBatching")
        self.gridLayoutBatching = QtWidgets.QGridLayout(self.groupBoxBatching)
        self.gridLayoutBatching.setObjectName("gridLayoutBatching")
        self.batchingWidgetPlaceholder = QtWidgets.QWidget(self.groupBoxBatching)
        self.batchingWidgetPlaceholder.setMinimumSize(QtCore.QSize(300, 300))
        self.batchingWidgetPlaceholder.setObjectName("batchingWidgetPlaceholder")
        self.gridLayoutBatching.addWidget(self.batchingWidgetPlaceholder, 0, 0, 1, 1)
        self.gridLayout_2.addWidget(self.groupBoxBatching, 1, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(
            20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
        )
        self.gridLayout_2.addItem(spacerItem, 2, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(
            40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum
        )
        self.horizontalLayout.addItem(spacerItem1)
        self.pushButtonCancel = QtWidgets.QPushButton(resamplingDialog)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout.addWidget(self.pushButtonCancel)
        self.pushButtonBatch = QtWidgets.QPushButton(resamplingDialog)
        self.pushButtonBatch.setObjectName("pushButtonBatch")
        self.horizontalLayout.addWidget(self.pushButtonBatch)
        self.pushButtonApply = QtWidgets.QPushButton(resamplingDialog)
        self.pushButtonApply.setObjectName("pushButtonApply")
        self.horizontalLayout.addWidget(self.pushButtonApply)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(resamplingDialog)
        self.pushButtonCancel.clicked.connect(resamplingDialog.reject)  # type: ignore
        self.pushButtonApply.clicked.connect(resamplingDialog.accept)  # type: ignore
        self.pushButtonBatch.clicked.connect(resamplingDialog.acceptBatch)  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(resamplingDialog)

    def retranslateUi(self, resamplingDialog):
        _translate = QtCore.QCoreApplication.translate
        resamplingDialog.setWindowTitle(
            _translate("resamplingDialog", "Meggie - Resampling")
        )
        self.groupBoxResample.setTitle(
            _translate("resamplingDialog", "Resampling options:")
        )
        self.labelCurrentRateHeading.setText(
            _translate("resamplingDialog", "Current rate:")
        )
        self.labelNewRateHeading.setText(_translate("resamplingDialog", "Resample to:"))
        self.groupBoxBatching.setTitle(_translate("resamplingDialog", "Batching"))
        self.pushButtonCancel.setText(_translate("resamplingDialog", "Cancel"))
        self.pushButtonBatch.setText(_translate("resamplingDialog", "Batch"))
        self.pushButtonApply.setText(_translate("resamplingDialog", "Apply"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './meggie/meggie/mainwindow/dialogs/aboutDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(921, 591)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(
            40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum
        )
        self.horizontalLayout_2.addItem(spacerItem)
        self.pushButtonClose = QtWidgets.QPushButton(Dialog)
        self.pushButtonClose.setObjectName("pushButtonClose")
        self.horizontalLayout_2.addWidget(self.pushButtonClose)
        self.gridLayout.addLayout(self.horizontalLayout_2, 2, 0, 1, 1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.labelVersion = QtWidgets.QLabel(Dialog)
        self.labelVersion.setObjectName("labelVersion")
        self.horizontalLayout.addWidget(self.labelVersion)
        spacerItem1 = QtWidgets.QSpacerItem(
            40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum
        )
        self.horizontalLayout.addItem(spacerItem1)
        self.lineEditVersion = QtWidgets.QLineEdit(Dialog)
        self.lineEditVersion.setMaxLength(100)
        self.lineEditVersion.setReadOnly(True)
        self.lineEditVersion.setObjectName("lineEditVersion")
        self.horizontalLayout.addWidget(self.lineEditVersion)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.textBrowser = QtWidgets.QTextBrowser(Dialog)
        self.textBrowser.setObjectName("textBrowser")
        self.verticalLayout.addWidget(self.textBrowser)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Meggie - About"))
        self.pushButtonClose.setText(_translate("Dialog", "Close"))
        self.labelVersion.setText(_translate("Dialog", "Current version:"))
        self.textBrowser.setHtml(
            _translate(
                "Dialog",
                '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">\n'
                '<html><head><meta name="qrichtext" content="1" /><style type="text/css">\n'
                "p, li { white-space: pre-wrap; }\n"
                "</style></head><body style=\" font-family:'Sans Serif'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;"><span style=" font-family:\'MS Shell Dlg 2\'; font-size:11pt;">Meggie is realeased under the Simplified BSD lisence:</span></p>\n'
                "<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'MS Shell Dlg 2'; font-size:11pt;\"><br /></p>\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">Copyright (c) &lt;2021&gt;, &lt;Kari Aliranta, Jaakko Leppäkangas, Janne Pesonen, Atte Rautio and CIBR&gt;</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">All rights reserved.</span></p>\n'
                "<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'monospace,Courier'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">Redistribution and use in source and binary forms, with or without</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">modification, are permitted provided that the following conditions are met: </span></p>\n'
                "<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'monospace,Courier'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">1. Redistributions of source code must retain the above copyright notice, this</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">   list of conditions and the following disclaimer. </span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">2. Redistributions in binary form must reproduce the above copyright notice,</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">   this list of conditions and the following disclaimer in the documentation</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">   and/or other materials provided with the distribution. </span></p>\n'
                "<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'monospace,Courier'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS &quot;AS IS&quot; AND</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.</span></p>\n'
                "<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'monospace,Courier'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">The views and conclusions contained in the software and documentation are those</span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">of the authors and should not be interpreted as representing official policies, </span></p>\n'
                '<p style=" margin-top:0px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;"><span style=" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;">either expressed or implied, of the FreeBSD Project.</span></p></body></html>',
            )
        )



"""Contains plot spectrum action handling."""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.channels import get_channels_by_type

from meggie.mainwindow.dynamic import Action

from meggie.actions.spectrum_plot.controller.spectrum import plot_spectrum_topo
from meggie.actions.spectrum_plot.controller.spectrum import plot_spectrum_averages

from meggie.utilities.dialogs.outputOptionsMain import OutputOptions


class PlotSpectrum(Action):
    """Plots spectrums."""

    def run(self, params={}):
        try:
            selected_name = self.data["outputs"]["spectrum"][0]
        except IndexError:
            return

        def option_handler(selected_option):
            params = {
                "name": selected_name,
                "output_option": selected_option,
                "channel_groups": self.experiment.channel_groups,
            }
            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = OutputOptions(self.window, handler=option_handler)
        dialog.show()

    def handler(self, subject, params):
        """ """
        if params["output_option"] == "channel_averages":
            plot_spectrum_averages(subject, params["channel_groups"], params["name"])
        else:
            info = subject.get_raw().info
            chs = list(get_channels_by_type(info).keys())
            if "eeg" in chs:
                plot_spectrum_topo(subject, params["name"], ch_type="eeg")
            if "grad" in chs or "mag" in chs:
                plot_spectrum_topo(subject, params["name"], ch_type="meg")



"""Contains implementation for raw resample."""

from meggie.utilities.threading import threaded

from meggie.actions.raw_resample.dialogs.resamplingDialogMain import ResamplingDialog

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class Resample(Action):
    """Shows a dialog for getting parameters and allows
    resampling data of the subject.
    """

    def run(self, params={}):
        resampling_dialog = ResamplingDialog(self.window, self.experiment, self.handler)
        resampling_dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """

        @threaded
        def resample_fun():
            raw = subject.get_raw()
            raw.resample(params["rate"])

        resample_fun(do_meanwhile=self.window.update_ui)
        subject.save()



"""Helper to start meggie, used by the setup.py script."""

import matplotlib
import joblib
import mne

import colorama

from meggie import mainWindowMain

# use Qt5Agg as TkAgg really hates threads
matplotlib.use("Qt5Agg")

# Use matplotlib as the backend for browser views for now,
# as it is difficult to bind to qt close events and
# because the functionality is still the same
mne.viz.set_browser_backend("matplotlib")

# Silence the failing of reset_all in colorama:
# RuntimeError: wrapped C/C++ object of type EmittingStream has been deleted
colorama.initialise.reset_all = lambda: None


# Joblib is so very verbose,
# and we have our own logging system,
# so silence it for now.
def patched_progress(self):
    return


joblib.parallel.Parallel.print_progress = patched_progress


def main():
    """Run Meggie main window."""
    mainWindowMain.main()


# To start meggie directly from the command-line
if __name__ == "__main__":
    mainWindowMain.main()



"""Contains implementation for evoked info"""

from meggie.mainwindow.dynamic import InfoAction


class Info(InfoAction):
    """Fills up evoked info box"""

    def run(self, params={}):
        try:
            selected_name = self.data["outputs"]["evoked"][0]
            evoked = self.experiment.active_subject.evoked[selected_name]
            params = evoked.params

            message = ""

            message += "Name: {}\n\n".format(evoked.name)

            if "conditions" in params:
                message += "Conditions: " + ", ".join(params["conditions"]) + "\n"

            if "groups" in params:
                for key, names in params["groups"].items():
                    message += "\nGroup " + str(key) + ": \n"
                    for name in names:
                        message += name + "\n"

        except Exception:
            message = ""

        return message



from meggie.utilities.testing import BaseTestAction
from meggie.actions.spectrum_info import Info


class TestSpectrumInfo(BaseTestAction):
    def test_spectrum_info(self):

        data = {"outputs": {"spectrum": ["Spectrum"]}}

        content = self.run_action(
            action_name="spectrum_info",
            handler=Info,
            data=data,
        )
        assert content



from meggie.utilities.testing import BaseTestAction
from meggie.actions.spectrum_delete import DeleteSpectrum


class TestSpectrumDelete(BaseTestAction):
    def test_spectrum_delete(self):

        data = {"outputs": {"spectrum": ["Spectrum"]}}

        self.run_action(
            action_name="spectrum_delete",
            handler=DeleteSpectrum,
            data=data,
        )



"""Contains a class for logic of action dialog."""

from PyQt5 import QtWidgets
import os
import json
import re
import logging

from meggie.mainwindow.dynamic import find_all_action_specs
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.filemanager import homepath
from meggie.utilities.serialization import deserialize_dict

from meggie.mainwindow.dialogs.actionDialogUi import Ui_ActionDialog


class ActionDialog(QtWidgets.QDialog):
    """Contains logic for action dialog."""

    def __init__(self, parent, experiment):
        QtWidgets.QDialog.__init__(self, parent)
        self.parent = parent
        self.experiment = experiment
        self.ui = Ui_ActionDialog()
        self.ui.setupUi(self)

        # clear the tree widget
        self.ui.treeWidgetActions.clear()

        try:
            # parse the actions log as a dict
            actions_dict = self._parse_log_as_dict()

            # transform to new dict that includes subject information
            self.data = self._transform_to_subjects(actions_dict)

            # populate the tree widget
            self._populate_tree_widget(self.data)
        except Exception as exc:
            exc_messagebox(self, exc)
            return

    def _parse_log_as_dict(self):
        """Parse actions log file into a dictionary."""
        experiment = self.experiment
        path = os.path.join(experiment.path, "actions.log")
        with open(path, "r") as f:
            lines = f.readlines()

        actions = {}

        for line in lines:
            content = json.loads(line)
            uid = content["uid"]

            if uid not in actions:
                actions[uid] = []

            actions[uid].append(content)

        subject_tree = {}

        for action_uid, action_rows in actions.items():
            action_tree = {}
            id_ = desc = version = data = None
            for row in action_rows:
                if row["type"] == "ACTION_START":
                    id_ = row["id"]
                    desc = row.get("desc")
                    version = row.get("version")
                    data = row.get("data")

                elif row["type"] == "SUBJECT_START":
                    subject_uid = row["subject_uid"]
                    if subject_uid not in action_tree:
                        action_tree[subject_uid] = {}
                    action_tree[subject_uid]["params"] = row["params"]

                elif row["type"] == "SUBJECT_END":
                    subject_uid = row["subject_uid"]
                    if subject_uid not in action_tree:
                        action_tree[subject_uid] = {}
                    action_tree[subject_uid]["finished"] = True
                    action_tree[subject_uid]["timestamp"] = row["timestamp"]

            for subject_uid, branch in action_tree.items():
                if branch.get("finished"):
                    branch["id"] = id_
                    branch["desc"] = desc
                    branch["data"] = data
                    branch["version"] = version
                    if subject_uid not in subject_tree:
                        subject_tree[subject_uid] = []
                    subject_tree[subject_uid].append(branch)

        return subject_tree

    def _transform_to_subjects(self, actions_data):
        """Inject subject data into the dict."""
        experiment = self.experiment

        subject_data = {}
        for subject_name, subject in experiment.subjects.items():
            if subject.uid in actions_data:
                subject_data[subject_name] = actions_data[subject.uid]

        return subject_data

    def _populate_tree_widget(self, data):
        """Populate the tree widget using the internal data representation."""

        action_specs = find_all_action_specs()

        for subject_name, subject_actions in sorted(data.items()):

            subject_item = QtWidgets.QTreeWidgetItem(self.ui.treeWidgetActions)
            subject_item.setText(0, subject_name)

            for action_data in subject_actions:
                action_item = QtWidgets.QTreeWidgetItem(subject_item)

                action_name = action_data["id"]
                if action_data["id"] in action_specs:
                    name_specs = action_specs[action_data["id"]][2]
                    action_name = name_specs.get("name", action_name)

                if action_data["desc"]:
                    action_name += f" ({action_data['desc']})"

                action_item.setText(0, action_name)

                params = action_data.get("params") or {}
                if params:
                    # for visual purposes, deserialize params to get correct types
                    deserialized_params = deserialize_dict(params)
                    params_item = QtWidgets.QTreeWidgetItem(action_item)
                    params_item.setText(0, "Params")
                    for param_name, param_value in deserialized_params.items():
                        try:
                            param_value = json.dumps(param_value)
                        except TypeError:
                            pass
                        msg = f"{param_name}: {param_value}"
                        param_item = QtWidgets.QTreeWidgetItem(params_item)
                        param_item.setText(0, msg)

                data = action_data.get("data") or {}
                if data:
                    data_item = QtWidgets.QTreeWidgetItem(action_item)
                    data_item.setText(0, "Data")
                    for data_name, data_value in data.items():
                        msg = f"{data_name}: {data_value}"
                        data_row_item = QtWidgets.QTreeWidgetItem(data_item)
                        data_row_item.setText(0, msg)

                timestamp_re = re.compile(
                    r"([0-9]*-[0-9]*-[0-9]*)T([0-9]*:[0-9]*:[0-9]*)\..*"
                )
                elems = timestamp_re.match(action_data["timestamp"])
                if elems:
                    msg = f"Timestamp: {elems[1]} {elems[2]} UTC"
                    timestamp_item = QtWidgets.QTreeWidgetItem(action_item)
                    timestamp_item.setText(0, msg)

    def on_treeWidgetActions_itemDoubleClicked(self, item):
        if item.childCount() == 0:
            # try creating a pretty formatted details dialog
            try:
                match = re.compile("([a-zA-Z0-9_]*): (.*)", re.DOTALL).match(
                    item.text(0)
                )
                if match:
                    val = match.group(2)
                    messagebox(self, str(val))
            except Exception:
                logging.getLogger("ui_logger").exception("")

    def on_pushButtonExport_clicked(self, checked=None):
        if checked is None:
            return

        default_filename = (
            f"{self.experiment.name.lower().replace(' ', '_')}_actions.json"
        )
        filepath, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Export Action Log",
            os.path.join(homepath(), default_filename),
            "JSON Files (*.json);;All Files (*)",
        )

        if filepath:
            try:
                with open(filepath, "w") as file:
                    json.dump(self.data, file, indent=4)
                logging.getLogger("ui_logger").info(
                    "Exported action log successfully to " + filepath
                )
            except Exception as exc:
                exc_messagebox(self, f"Failed to export: {exc}")

        self.close()



from PyQt5 import QtWidgets

from meggie.utilities.testing import BaseTestAction
from meggie.actions.epochs_create import CreateEpochs
from meggie.actions.epochs_create.dialogs.createEpochsFromEventsDialogMain import (
    CreateEpochsFromEventsDialog,
)


class TestEpochsCreate(BaseTestAction):
    def test_epochs_create(self):

        self.run_action(
            action_name="epochs_create",
            handler=CreateEpochs,
            patch_paths=[
                "meggie.actions.epochs_create.dialogs.createEpochsFromEventsDialogMain"
            ],
        )
        dialog = self.find_dialog(CreateEpochsFromEventsDialog)

        event = {"event_id": 1, "mask": 0}
        item = QtWidgets.QListWidgetItem(
            "%s, %s" % ("ID " + str(event["event_id"]), "mask=" + str(event["mask"]))
        )
        dialog.ui.listWidgetEvents.addItem(item)
        dialog.events = [event]

        dialog.accept()



"""Contains a class for handling reading and storing of
global preferences such as the previous experiment and
the tab settings.
"""

import os
import configparser
import logging

from meggie.utilities.filemanager import configpath
from meggie.utilities.filemanager import datapath


class PreferencesHandler(object):
    """Class for storing and setting preferences."""

    def __init__(self, prefs_path=""):

        if prefs_path:
            self.prefs_path = prefs_path
        else:
            self.prefs_path = os.path.join(configpath(), "preferences.cfg")

        self.workspace = datapath()

        self.previous_experiment_name = ""
        self.auto_load_last_open_experiment = False
        self.active_plugins = []

        self.read_preferences_from_disk()

    def read_config(self):
        """Reads the config file from file system"""
        config = configparser.RawConfigParser()
        if os.path.isfile(self.prefs_path):
            config.read(self.prefs_path)
        return config

    def write_preferences_to_disk(self):
        """Writes the preferences to file system, in INI style."""
        # Base the new config on the old one.
        config = self.read_config()
        try:
            config.add_section("MiscOptions")
            config.add_section("Workspace")
        except configparser.DuplicateSectionError:
            pass

        # Sanity of these values is assumed to be checked by the calling method

        config.set("Workspace", "workspaceDir", self.workspace)
        logging.getLogger("ui_logger").info("Workspace: " + self.workspace)

        config.set(
            "MiscOptions", "previousExperimentName", self.previous_experiment_name
        )
        logging.getLogger("ui_logger").info(
            "Previous experiment: " + self.previous_experiment_name
        )

        if self.auto_load_last_open_experiment:
            config.set("MiscOptions", "autoReloadPreviousExperiment", "True")
        else:
            config.set("MiscOptions", "autoReloadPreviousExperiment", "False")
        logging.getLogger("ui_logger").info(
            "Auto reload previous experiment: "
            + str(self.auto_load_last_open_experiment)
        )

        config.set("MiscOptions", "activePlugins", ",".join(self.active_plugins))
        logging.getLogger("ui_logger").info(
            "Active plugins: " + str(self.active_plugins)
        )

        try:
            os.makedirs(os.path.dirname(self.prefs_path))
        except PermissionError:
            logging.getLogger("ui_logger").exception(
                "Could not save the configuration file."
            )
        except FileExistsError:
            pass

        with open(self.prefs_path, "w") as configfile:
            config.write(configfile)

    def read_preferences_from_disk(self):
        """Reads the preferences from file system into attributes."""

        logging.getLogger("ui_logger").info(
            f"Reading preferences from: {self.prefs_path}"
        )

        config = self.read_config()

        try:
            self.workspace = config.get("Workspace", "workspaceDir")
        except Exception:
            pass

        try:
            auto_reload = config.get("MiscOptions", "autoReloadPreviousExperiment")
            if auto_reload == "True":
                self.auto_load_last_open_experiment = True
        except Exception:
            pass

        try:
            self.previous_experiment_name = config.get(
                "MiscOptions", "previousExperimentName"
            )
        except Exception:
            pass

        try:
            active_plugins = config.get("MiscOptions", "activePlugins")
            self.active_plugins = active_plugins.split(",")
        except Exception:
            pass



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_plot_projections import PlotProjections


class TestRawPlotProjections(BaseTestAction):
    def test_raw_plot_projections(self):

        self.run_action(
            action_name="raw_plot_projections",
            handler=PlotProjections,
        )



"""Helpers to handle data units. Data in memory is in standard units, i.e
for magnetometers it is T, gradiometers it is T/m, and electrodes it is V.
However it is customary to convert these into fT, fT/cm and uV respectively
for human reading. For power densities it follows that the units will be
T^2/Hz, (T/m)^2/Hz and V^2/Hz. This is because density is the FT of the
autocorrelation of the signal, and thus if the original unit is X,
the resulting unit is X^2/Hz. Or if the logarithmic transformation is done,
unit is dB/Hz for each.)
"""

UNITS = {
    "mag": {
        "scaling": 1e15,
        "unit": "fT",
        "power_unit": "T^2/Hz",
    },
    "grad": {
        "scaling": 1e13,
        "unit": "fT/cm",
        "power_unit": "(T/m)^2/Hz",
    },
    "eeg": {"scaling": 1e6, "unit": "uV", "power_unit": "V^2/Hz"},
    "eog": {"scaling": 1e6, "unit": "uV", "power_unit": "V^2/Hz"},
    "ecg": {"scaling": 1e6, "unit": "uV", "power_unit": "V^2/Hz"},
}


def get_scaling(type_):
    """Returns standard scaling factor for a sensor type.

    Parameters
    ----------
    type_ : str
        The sensor type.

    Returns
    -------
    float
        The scaling factor.
    """
    if type_ not in UNITS:
        raise Exception("Unknown data type")
    return UNITS[type_]["scaling"]


def get_unit(type_):
    """Returns unit of signal for a sensor type.

    Parameters
    ----------
    type_ : str
        The sensor type.

    Returns
    -------
    str
        The unit.
    """
    if type_ not in UNITS:
        raise Exception("Unknown data type")
    return UNITS[type_]["unit"]


def get_power_unit(type_, log=False):
    """Returns unit of power for a sensor type.

    Parameters
    ----------
    type_ : str
        The sensor type.
    log : bool
        Whether the unit is for log transformed data.

    Returns
    -------
    str
        The unit.
    """
    if type_ not in UNITS:
        raise Exception("Unknown data type")
    if log:
        return "dB/Hz"
    else:
        return UNITS[type_]["power_unit"]



from meggie.utilities.testing import BaseTestAction
from meggie.actions.spectrum_create import CreateSpectrum
from meggie.utilities.dialogs.powerSpectrumDialogMain import (
    PowerSpectrumDialog,
)


class TestSpectrumCreate(BaseTestAction):
    def test_spectrum_create(self):

        self.run_action(
            action_name="spectrum_create",
            handler=CreateSpectrum,
            patch_paths=["meggie.utilities.dialogs.powerSpectrumDialogMain"],
        )
        raw = self.experiment.active_subject.get_raw()
        dialog = self.find_dialog(PowerSpectrumDialog)
        dialog.add_intervals([("fixed", ("1", 0.0, raw.times[-1]))])
        dialog.accept()



"""Contains implementation for epochs info"""

from meggie.mainwindow.dynamic import InfoAction

from meggie.utilities.formats import format_float


class Info(InfoAction):
    """Fills up epochs info box"""

    def run(self, params={}):

        try:
            selected_name = self.data["outputs"]["epochs"][0]

            epochs = self.experiment.active_subject.epochs[selected_name]
            params = epochs.params

            message = ""

            message += "Name: {0}\n\n".format(epochs.name)

            message += "Count: {0}\n".format(epochs.count)

            message += "Baseline: {0}s - {1}s\n".format(
                format_float(params["bstart"]), format_float(params["bend"])
            )
            message += "Times: {0}s - {1}s\n".format(
                format_float(params["tmin"]), format_float(params["tmax"])
            )

            message += "\nReject params: \n"
            if "grad" in params:
                message += "Gradiometers: {}\n".format(params.get("reject").get("grad"))
            if "mag" in params:
                message += "Magnetometers: {}\n".format(params.get("reject").get("mag"))
            if "eeg" in params:
                message += "EEG: {}\n".format(params.get("reject").get("eeg"))

            if "events" in params:
                message += "\nCreated from events: \n"
                for event in params.get("events"):
                    message += "ID: {0}, mask: {1}\n".format(
                        event["event_id"], event["mask"]
                    )

        except Exception:
            message = ""

        return message



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_event_info import Info


class TestRawEventInfo(BaseTestAction):
    def test_raw_event_info(self):

        content = self.run_action(
            action_name="raw_event_info",
            handler=Info,
        )
        assert content



"""Contains a class for logic of the rereferencing dialog."""

import logging

import mne

from PyQt5 import QtWidgets

from meggie.actions.raw_rereference.dialogs.rereferencingDialogUi import (
    Ui_rereferencingDialog,
)

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget
from meggie.utilities.messaging import exc_messagebox


class RereferencingDialog(QtWidgets.QDialog):
    """Contains logic for the rereferencing dialog."""

    def __init__(self, parent, experiment, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_rereferencingDialog()
        self.ui.setupUi(self)

        self.experiment = experiment
        self.parent = parent
        self.handler = handler

        subject = self.experiment.active_subject
        raw = subject.get_raw()

        # fill the combobox
        picks = mne.pick_types(raw.info, eeg=True, meg=False, eog=True)
        ch_names = [
            ch_name
            for ch_idx, ch_name in enumerate(raw.info["ch_names"])
            if ch_idx in picks
        ]

        self.ui.listWidgetChannels.clear()
        for ch_name in ch_names:
            self.ui.listWidgetChannels.addItem(ch_name)

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry(),
        )
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

    def _experiment_getter(self):
        return self.experiment

    def accept(self):
        experiment = self.experiment
        subject = experiment.active_subject

        if self.ui.radioButtonUseAverage.isChecked():
            selection = "average"
        else:
            selection = [
                item.text() for item in self.ui.listWidgetChannels.selectedItems()
            ]

        try:
            params = {"selection": selection}
            self.handler(subject, params)
        except Exception as exc:
            exc_messagebox(self.parent, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self):
        experiment = self.experiment

        if self.ui.radioButtonUseAverage.isChecked():
            selection = "average"
        else:
            selection = [
                item.text() for item in self.ui.listWidgetChannels.selectedItems()
            ]

        params = {"selection": selection, "run_in_batch": True}

        selected_subject_names = self.batching_widget.selected_subjects
        for name, subject in experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append((subject, str(exc)))
                    logging.getLogger("ui_logger").exception("")

        self.batching_widget.cleanup()

        self.parent.initialize_ui()

        self.close()



"""Contains controlling logic for the evoked implementation."""

import logging

import mne

from meggie.datatypes.evoked.evoked import Evoked

from meggie.utilities.validators import assert_arrays_same

from meggie.utilities.threading import threaded


@threaded
def group_average_evoked(experiment, evoked_name, groups, new_name):
    """Computes group average item."""
    keys = []
    time_arrays = []
    for group_key, group_subjects in groups.items():
        for subject_name in group_subjects:
            try:
                subject = experiment.subjects.get(subject_name)
                evoked = subject.evoked.get(evoked_name)

                mne_evokeds = evoked.content
                for mne_evoked in mne_evokeds.values():
                    time_arrays.append(mne_evoked.times)
            except Exception:
                continue

    assert_arrays_same(time_arrays)

    # handle channel differences
    ch_names = []
    for group_key, group_subjects in groups.items():
        for subject_name in group_subjects:
            try:
                subject = experiment.subjects.get(subject_name)
                evoked = subject.evoked.get(evoked_name)

                mne_evokeds = evoked.content
                for mne_evoked in mne_evokeds.values():
                    ch_idxs = mne.pick_types(mne_evoked.info, meg=True, eeg=True)
                    ch_names.append(
                        tuple(
                            [
                                ch_name
                                for ch_idx, ch_name in enumerate(
                                    mne_evoked.info["ch_names"]
                                )
                                if ch_idx in ch_idxs
                            ]
                        )
                    )
            except Exception:
                continue

    if len(set(ch_names)) != 1:
        logging.getLogger("ui_logger").debug(
            "Evokeds contain different sets of good channels. Identifying common ones.."
        )

        common_ch_names = list(set.intersection(*map(set, ch_names)))

        logging.getLogger("ui_logger").debug(
            str(len(common_ch_names)) + " common channels found."
        )
    else:
        common_ch_names = ch_names[0]

    grand_evokeds = {}
    for group_key, group_subjects in groups.items():
        for subject in experiment.subjects.values():
            if subject.name not in group_subjects:
                continue

            evoked = subject.evoked.get(evoked_name)
            if not evoked:
                logging.getLogger("ui_logger").warning(
                    evoked_name + " not present for " + subject.name
                )
                continue

            for evoked_item_key, evoked_item in evoked.content.items():
                grand_key = (group_key, evoked_item_key)

                if grand_key not in grand_evokeds:
                    grand_evokeds[grand_key] = []
                grand_evokeds[grand_key].append(
                    evoked_item.copy().pick(common_ch_names)
                )

    grand_averages = {}
    new_keys = []
    for key, grand_evoked in grand_evokeds.items():
        new_key = str(key[1]) + "_" + str(key[0])
        if len(grand_evoked) == 1:
            grand_averages[new_key] = grand_evoked[0].copy()
        else:
            grand_averages[new_key] = mne.grand_average(grand_evoked)
        new_keys.append(new_key)

    # to avoid mne complaints, set dev_head_t manually to same for all
    keys = sorted(grand_averages.keys())
    for key in keys:
        grand_averages[key].info["dev_head_t"] = grand_averages[keys[0]].info[
            "dev_head_t"
        ]

    subject = experiment.active_subject
    evoked_directory = subject.evoked_directory
    params = {"conditions": new_keys, "groups": groups}

    grand_average_evoked = Evoked(
        new_name, evoked_directory, params, content=grand_averages
    )

    grand_average_evoked.save_content()
    subject.add(grand_average_evoked, "evoked")



from meggie.utilities.testing import BaseTestAction
from meggie.actions.evoked_plot_topomap import PlotEvokedTopomap
from meggie.actions.evoked_plot_topomap.dialogs.evokedTopomapDialogMain import (
    EvokedTopomapDialog,
)


class TestEvokedPlotTopomap(BaseTestAction):
    def test_evoked_plot_topomap(self):

        data = {"outputs": {"evoked": ["Evoked"]}}

        self.run_action(
            action_name="evoked_plot_topomap",
            handler=PlotEvokedTopomap,
            data=data,
            patch_paths=[
                "meggie.actions.evoked_plot_topomap.dialogs.evokedTopomapDialogMain"
            ],
        )
        dialog = self.find_dialog(EvokedTopomapDialog)
        dialog.accept()



"""Contains implementation for tfr info"""

from meggie.mainwindow.dynamic import InfoAction

from meggie.utilities.formats import format_float
from meggie.utilities.formats import format_floats


class Info(InfoAction):
    """Fills up tfr info box"""

    def run(self, params={}):
        try:
            selected_name = self.data["outputs"]["tfr"][0]

            tfr = self.experiment.active_subject.tfr[selected_name]
            params = tfr.params

            message = ""
            if "decim" in params:
                message += "Decimated by factor: {0}\n".format(params["decim"])
            if "evoked_subtracted" in params:
                message += "Evoked subtracted: {0}\n".format(
                    params["evoked_subtracted"]
                )
            if "conditions" in params:
                message += "Conditions: " + ", ".join(params["conditions"]) + "\n"
            if "n_cycles" in params:
                if hasattr(params["n_cycles"], "__len__"):
                    message += (
                        "Cycles (per frequency): "
                        + ", ".join(format_floats(params["n_cycles"]))
                        + "\n"
                    )
                else:
                    message += f"Cycles (fixed): {params['n_cycles']}\n"

            if hasattr(tfr, "times"):
                message += "Times: {0}s - {1}s\n".format(
                    format_float(tfr.times[0]), format_float(tfr.times[-1])
                )

            if hasattr(tfr, "freqs"):
                message += "Freqs: {0} - {1}\n".format(
                    format_float(tfr.freqs[0]), format_float(tfr.freqs[-1])
                )

            if "groups" in params:
                for key, names in params["groups"].items():
                    message += "\nGroup " + str(key) + ": \n"
                    for name in names:
                        message += name + "\n"

        except Exception:
            message = ""

        return message



from meggie.utilities.testing import BaseTestAction
from meggie.actions.epochs_save import SaveEpochs
from meggie.utilities.dialogs.outputOptionsMain import OutputOptions


class TestEpochsSaveAllChannels(BaseTestAction):
    def test_epochs_save_all_channels(self):

        data = {"outputs": {"epochs": ["Epochs"]}}

        self.run_action(
            action_name="epochs_save",
            handler=SaveEpochs,
            data=data,
        )
        dialog = self.find_dialog(OutputOptions)
        dialog.ui.radioButtonChannelAverages.setChecked(False)
        dialog.accept()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.epochs_plot_image import PlotEpochsImage


class TestEpochsPlotImage(BaseTestAction):
    def test_epochs_plot_image(self):

        data = {"outputs": {"epochs": ["Epochs"]}}

        self.run_action(
            action_name="epochs_plot_image",
            handler=PlotEpochsImage,
            data=data,
        )



import os

import mne
import matplotlib.pyplot as plt

from meggie.utilities.channels import get_triplet_from_mag
from meggie.utilities.channels import get_channels_by_type
from meggie.utilities.channels import get_default_channel_groups
from meggie.utilities.channels import clean_names
from meggie.utilities.channels import iterate_topography
from meggie.utilities.channels import average_to_channel_groups


def test_get_triplet_from_mag():
    assert get_triplet_from_mag("MEG2031") == ["MEG2031", "MEG2032", "MEG2033"]


def test_get_channels_by_type():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, "MEG", "sample", "sample_audvis_raw.fif")
    info = mne.io.read_raw_fif(sample_fname).info

    chs = get_channels_by_type(info)
    assert set(chs.keys()) == set(["grad", "mag", "eeg"])
    assert len(chs["eeg"]) == 59
    assert len(chs["grad"]) == 203
    assert len(chs["mag"]) == 102


def test_default_channel_groups():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, "MEG", "sample", "sample_audvis_raw.fif")

    raw = mne.io.read_raw_fif(sample_fname)

    assert get_default_channel_groups(raw.info, "eeg")["Left-frontal"] == [
        "EEG 001",
        "EEG 004",
        "EEG 005",
        "EEG 009",
        "EEG 010",
        "EEG 011",
        "EEG 012",
    ]


def test_clean_names():
    assert clean_names(["EEG 012"]) == clean_names(["EEG012"]) == ["EEG012"]


def test_iterate_topography():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, "MEG", "sample", "sample_audvis_raw.fif")
    info = mne.io.read_raw_fif(sample_fname).info

    fig = plt.figure()

    ch_names = info["ch_names"][::2]

    def on_pick(ax, info_idx, names_idx):
        pass

    iterated = []
    for ax, info_idx, names_idx in iterate_topography(fig, info, ch_names, on_pick):
        iterated.append((info_idx, names_idx))

    assert iterated[0] == (0, 0)
    assert iterated[-1] == (304, 152)


def test_average_to_channel_groups():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, "MEG", "sample", "sample_audvis_raw.fif")

    raw = mne.io.read_raw_fif(sample_fname, preload=True)
    info = raw.info

    ch_names = info["ch_names"][:20]

    meg_channel_groups = get_default_channel_groups(raw.info, "meg")
    eeg_channel_groups = get_default_channel_groups(raw.info, "eeg")

    # find out to which channel group each of the channels belongs to
    results = []
    for ch_name in ch_names:
        for ch_group_name, ch_group in meg_channel_groups.items():
            if ch_name in ch_group:
                results.append((ch_name, ch_group_name))

    labels, averaged_data = average_to_channel_groups(
        raw._data[:, 0:100],
        info,
        ch_names,
        {"meg": meg_channel_groups, "eeg": eeg_channel_groups},
    )

    # check that the localization of data is same before and after averaging
    assert set([result[1] for result in results]) == set([label[1] for label in labels])



from meggie.utilities.testing import BaseTestAction
from meggie.actions.tfr_create import CreateTFR
from meggie.actions.tfr_create.dialogs.TFRDialogMain import TFRDialog


class TestTFRCreate(BaseTestAction):
    def test_tfr_create(self):

        data = {"inputs": {"epochs": ["Epochs"]}}

        self.run_action(
            action_name="tfr_create",
            handler=CreateTFR,
            data=data,
            patch_paths=["meggie.actions.tfr_create.dialogs.TFRDialogMain"],
        )
        dialog = self.find_dialog(TFRDialog)
        dialog.ui.radioButtonFixed.setChecked(True)
        dialog.ui.doubleSpinBoxMinFreq.setValue(20)
        dialog.ui.doubleSpinBoxMaxFreq.setValue(40)
        dialog.ui.doubleSpinBoxNcycles.setValue(1.0)
        dialog.accept()



"""Contains implementation for evoked topomaps"""

import matplotlib.pyplot as plt
import numpy as np

from matplotlib.gridspec import GridSpec

from meggie.utilities.channels import get_channels_by_type
from meggie.utilities.plotting import set_figure_title

from meggie.mainwindow.dynamic import Action

from meggie.actions.evoked_plot_topomap.dialogs.evokedTopomapDialogMain import (
    EvokedTopomapDialog,
)


class PlotEvokedTopomap(Action):
    """Plots a sequence of topomaps."""

    def run(self, params={}):

        try:
            selected_name = self.data["outputs"]["evoked"][0]
        except IndexError:
            return

        subject = self.experiment.active_subject
        evoked = subject.evoked.get(selected_name)

        def handle_close(tmin, tmax, step, radius):
            params = {}
            params["tmin"] = tmin
            params["tmax"] = tmax
            params["step"] = step
            params["radius"] = radius
            params["name"] = selected_name

            self.handler(subject, params)

        dialog = EvokedTopomapDialog(self.window, evoked, handle_close)
        dialog.show()

    def handler(self, subject, params):
        name = params["name"]
        tmin = params["tmin"]
        tmax = params["tmax"]
        step = params["step"]
        radius = params["radius"]
        evoked = subject.evoked.get(name)

        for key, mne_evoked in sorted(evoked.content.items()):
            channels = get_channels_by_type(mne_evoked.info)
            for ch_type in channels.keys():
                title_elems = [name, key, ch_type]
                times = np.arange(tmin, tmax, step)

                # Use custom figure so that mne does not remove the mpl toolbar
                fig = plt.figure()
                axes = []

                # one subplot for each topomap
                for idx in range(len(times)):
                    spec = GridSpec(5, 2 * (len(times) + 1)).new_subplotspec(
                        (1, 2 * idx), rowspan=3, colspan=2
                    )
                    axes.append(fig.add_subplot(spec))

                # and one for colorbar
                spec = GridSpec(5, 2 * (len(times) + 1)).new_subplotspec(
                    (1, 2 * len(times)), rowspan=3, colspan=1
                )
                axes.append(fig.add_subplot(spec))

                # until there is a good solution to topomap skirts, fix a value
                sphere = None
                if ch_type in ["mag", "grad"]:
                    sphere = radius

                fig = mne_evoked.plot_topomap(
                    times=times, ch_type=ch_type, axes=axes, sphere=sphere
                )
                set_figure_title(fig, "_".join(title_elems))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './meggie/meggie/utilities/dialogs/shortQuestionBoxUi.ui'
#
# Created by: PyQt5 UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_shortQuestionBox(object):
    def setupUi(self, shortQuestionBox):
        shortQuestionBox.setObjectName("shortQuestionBox")
        shortQuestionBox.resize(582, 117)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(shortQuestionBox.sizePolicy().hasHeightForWidth())
        shortQuestionBox.setSizePolicy(sizePolicy)
        self.gridLayout = QtWidgets.QGridLayout(shortQuestionBox)
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(
            40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum
        )
        self.gridLayout.addItem(spacerItem, 1, 0, 1, 1)
        self.pushButtonCancel = QtWidgets.QPushButton(shortQuestionBox)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.gridLayout.addWidget(self.pushButtonCancel, 1, 1, 1, 1)
        self.pushButtonOk = QtWidgets.QPushButton(shortQuestionBox)
        self.pushButtonOk.setObjectName("pushButtonOk")
        self.gridLayout.addWidget(self.pushButtonOk, 1, 2, 1, 1)
        self.scrollArea = QtWidgets.QScrollArea(shortQuestionBox)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 562, 68))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.labelMessage = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelMessage.sizePolicy().hasHeightForWidth())
        self.labelMessage.setSizePolicy(sizePolicy)
        self.labelMessage.setAlignment(
            QtCore.Qt.AlignLeading | QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter
        )
        self.labelMessage.setWordWrap(True)
        self.labelMessage.setObjectName("labelMessage")
        self.gridLayout_2.addWidget(self.labelMessage, 0, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 0, 0, 1, 3)

        self.retranslateUi(shortQuestionBox)
        self.pushButtonOk.clicked.connect(shortQuestionBox.accept)  # type: ignore
        self.pushButtonCancel.clicked.connect(shortQuestionBox.reject)  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(shortQuestionBox)

    def retranslateUi(self, shortQuestionBox):
        _translate = QtCore.QCoreApplication.translate
        shortQuestionBox.setWindowTitle(_translate("shortQuestionBox", "Meggie"))
        self.pushButtonCancel.setText(_translate("shortQuestionBox", "Cancel"))
        self.pushButtonOk.setText(_translate("shortQuestionBox", "Ok"))
        self.labelMessage.setText(
            _translate(
                "shortQuestionBox",
                "The was a cat behind the all mighty bird singing hallelujah. Who knew it would be so devastating if not all of us? The cries in the woods, howls of the never ending human culture vanishing into thin q numbers.",
            )
        )
